#! /usr/bin/python3

"""
Run tests on two different branches of the same repo and compare the
results.
Usage:
nostrils [-f|--force] [previous [current]] [-- FILE1 [...]]

If "previous" is not given, it defaults to "master". If "current" is not given,
it defaults to the current checked out commit.

"--force" means that tests will be rerun even if there already were cached
results for one or both of the two commits.

Any uncommitted change in the working tree is neglected.

When done, calls any executable file in ~/config/nostrils/alarms.
"""

__version__ = '0.2'

import shutil
import tempfile
import datetime
import subprocess
from os import environ, listdir, mkdir
from os.path import join, exists, realpath, basename
from os import makedirs, chdir, remove
from subprocess import call, Popen, DEVNULL
from time import sleep
import argparse
import sys
import ast

from xdg.BaseDirectory import xdg_cache_home, xdg_config_home
from git import Repo

DESC = 'Compare test runs across git commits.'
parser = argparse.ArgumentParser(description=DESC)
parser.add_argument("current", nargs='?', default=None)
parser.add_argument("other", nargs='?', default='master')
parser.add_argument("files", nargs=argparse.REMAINDER,
                    help="Run on the given files, which must be preceded by "
                         "\"--\", as in:\n"
                         "\"nostrils.py mybranch master -- tests/mytest.py\". "
                         "Implies -F, requires \"other\" and "
                         "\"current\" to be explicitly given.")

# Not implemented:
#parser.add_argument("-f", "--force", dest='forced_hash', nargs="?",
#                    default=set(),
#                    help="Rerun test for given hash even if old results are "
#                         "present for it.")
parser.add_argument("-F", "--force-all", dest='force', action='store_true',
                    help="Run tests for all hashes (do not look for cached "
                         "results).")
parser.add_argument("-c", "--copy", dest='copy', action='store_true',
                    help="Use tests from \"current\" also to test \"other\". "
                    "Valid only in conjuction with --only, implies -F.")
parser.add_argument("-s", "--silent", dest='silent', action='store_true',
                    help="Do not run any alarms at the end of tests (if any).")


args = parser.parse_args()

test_commands_path = join(xdg_config_home, 'nostrils', 'test_commands.py')
if exists(test_commands_path):
    with open(test_commands_path) as fin:
        test_commands = ast.literal_eval(fin.read())
else:
    test_commands = {}

cache_dir = join(xdg_cache_home, 'nostrils')

class TestCommand:
    def __init__(self):
        self._command = environ.get('TESTCMD',
                                    test_commands.get(realpath('.'),
                                                      'pytest'))

    def _filesystem_repr(self):
        return self._command.replace('/', '_-_')

    def results_dir(self):
        if args.files:
            res_dir = join(cache_dir, 'partial', self._filesystem_repr())
        else:
            res_dir = join(cache_dir, 'results', self._filesystem_repr())

        if not exists(res_dir):
            makedirs(res_dir)
        return res_dir

    def for_subprocess(self):
        return [self._command] + args.files

    def __str__(self):
        return self._command

class Commit(str):
    def path_out(self, test_command):
        return join(test_command.results_dir(), self)

def build_exts(interpreter, dirs):
    print("Building extensions with %s" % interpreter)
    pops = []
    for a_dir in dirs:
        chdir(a_dir)
        pops.append(Popen([interpreter, 'setup.py', 'build_ext', '--inplace'],
                          stderr=DEVNULL, stdout=DEVNULL))
        chdir('..')
    while not all([p.poll() != None for p in pops]):
        sleep(.1)
    print("Finished building extensions")

def run_tests(dirs, hashes, origpath, test_command):
    pops = []
    print("Running tests on", hashes)
    fouts = [None, None]
    tempdirs = []
    for idx in range(len(dirs)):
        chdir(dirs[idx])
        fout = tempfile.NamedTemporaryFile(delete=False)
        fouts[idx] = fout
        tempdir = tempfile.mkdtemp()
        tempdirs.append(tempdir)
        # Run with ad-hoc temporary directory, to avoid the two different
        # "pytest" processes to mess with files in '/tmp' (the environmental
        # variable is respect by the "tempfile" module)
        pops.append(Popen(test_command.for_subprocess(),
                          stderr=fout, stdout=fout,
                          env=dict(TMPDIR=tempdir, **environ)))
        chdir('..')
    while not all([p.poll() != None for p in pops]):
        sleep(.1)
    print("Finished running tests")
    for idx, fout in enumerate(fouts):
        if fout is None:
            continue
        fout.close()
        # We store even when we ran only part of the tests - but that's only
        # for the diff, it's not reused:
        with open(hashes[idx].path_out(test_command), 'w') as fdef:
            with open(fout.name) as ftemp:
                fdef.write(ftemp.read().replace(dirs[idx], origpath))
        remove(fout.name)
    for tempdir in tempdirs:
        shutil.rmtree(tempdir)

def print_diff(outfiles):
    call(['git', 'diff', '--no-index', '--color-words'] + outfiles)

def clean_dirs(dirs):
    for a_dir in dirs:
        shutil.rmtree(a_dir)

def run_alarms():
    warners_dir = join(xdg_config_home, 'nostrils', 'alarms')
    for warner in listdir(warners_dir):
        full_path = join(warners_dir, warner)
        try:
            subprocess.Popen([full_path])
        except Exception as exc:
            print("\nFailed running alarm %s (see error below)\n" % full_path)
            raise

def main():
    
    repo = Repo('.')
    refs = args.other, args.current
    if ('--' in sys.argv and
        len(sys.argv) - sys.argv.index('--') - 1 != len(args.files)):
        # FIXME!
        # ArgParse suppresses the "--" from the arguments list ONLY IF previous
        # positional arguments are present.
        raise NotImplementedError("You need to explicitly provide the two "
                                  "references when selecting files to test.")
    test_command = TestCommand()
    print("Nostriling \"%s\" on %s at %s" % (test_command, refs, str(datetime.datetime.now())))
    hashes = [Commit(repo.commit(ref).hexsha) for ref in refs]
    print(hashes)
    assert(hashes[0] != hashes[1]), "The two given references coincide."
    todo = []
    if args.force or args.files:
        todo = list(hashes)
    else:
        for a_hash in hashes:
            if exists(a_hash.path_out(test_command)):
                print(a_hash, "does not need to be retested")
            else:
                todo.append(a_hash)

    if todo:
        refs_map = {repo.commit(ref).hexsha : ref for ref in refs}
        repo_path = realpath('.')
        repo_name = basename(repo_path)
        chdir(cache_dir)
        dirnames = []
        checked_out = repo.commit().hexsha
        for a_hash in todo:
            print("Preparing", a_hash)
            dirname = realpath("%s_%s" % (repo_name, a_hash))
            dirnames.append(dirname)
            if exists(dirname):
                # A previous run was aborted or failed:
                shutil.rmtree(dirname)
            shutil.copytree(repo_path, dirname)
            print("Move to", dirname)
            chdir(dirname)
            print("Moved to", dirname)
            if a_hash != checked_out:
                print("Checkout", a_hash)
                call(['git', 'checkout', refs_map[a_hash]])
                print("Checked out", refs_map[a_hash])
            if args.copy and a_hash == hashes[0]:
                if not args.files:
                    raise ValueError("Test files must be explicitly passed to "
                                     "use option -c.")
                call(['git', 'checkout', hashes[1], '--'] + args.files)
                print("Checked out %s from %s" % (args.files, hashes[1]))
#            call(['git', 'reset', '--hard'])
            chdir("..")

    
        build_exts('python', dirnames)
        build_exts('python3', dirnames)
        try:
            run_tests(dirnames, todo, repo_path, test_command)
        finally:
            clean_dirs(dirnames)

    if todo and not args.silent:
        run_alarms()
    print_diff([h.path_out(test_command) for h in hashes])

if __name__ == '__main__':
    main()
